# Author: Bas Terwijn
import re

class ProteinFoldException(Exception):
    def __init__(self, message):
        super(Exception, self).__init__(message)

class Coordinate:
    """N-dimensional Coordinate representattion"""

    def __init__(self,N):
        """All zero N-dimensional Coordinate"""
        self.coord=[]
        for i in range(N):
            self.coord.append(0)

    def __str__(self):
        """Returns string representation of a Coordinate"""
        return str(self.coord)

    def fold(self,fold):
        """Changes abs(fold) element of Coordinate.
        If fold is positive the element is increased by 1.
        If negative the element is decreased by 1"""
        index=abs(fold)
        self.coord[index-1]+= (1 if fold>0 else -1)


class Protein:
    """Representation of a folded Protein for which we can compute the score"""

    def __init__(self,filename):
        """Empty Protein"""
        self.elements=[]
        self.folds=[]
        self.dimension=0
        self.load(filename)

    def __str__(self):
        """Returns string representation of a Protein"""
        s=""
        s+="  elements    : "
        for i in self.elements:
            s+=i
        s+="\n"
        s+="  nr elements : "+str(len(self.elements))+"\n"
        s+="  folds       : "+str(self.folds)+"\n"
        s+="  dimension   : "+str(self.dimension)
        return s

    def load(self,filename):
        """Loads Protein from file"""
        line_count=0
        with open(filename,"r") as f:
            for line in f:
                line=re.sub(r"[\n\t\s]*", "", line) # remove newlines,tabs,whitespaces
                if len(line)>0 and line[0]!='#':
                    if line_count==0:
                        for i in line:
                            self.elements.append(i)
                    elif line_count==1:
                        folds=line.split(",")
                        for i in folds:
                            fold=int(i)
                            if fold!=0:
                                self.folds.append(fold)
                                if abs(fold)>self.dimension:
                                    self.dimension=abs(fold)
                            else:
                                raise ProteinFoldException("fold ("+str(i)+") may not be zero.")
                    else:
                        print("ignoring line: "+line)
                    line_count+=1
        if not (len(self.folds)==len(self.elements)-1 or len(self.folds)==0):
            raise ProteinFoldException("The number of folds in '"+filename+"' should be 1 less than the number of Protein elements, so there should be "+str(len(self.elements)-1)+" comma separated folds instead of currently: "+str(len(self.folds)))

    def get_elements(self):
        return self.elements

    def get_dimension(self):
        return self.dimension

    def make_grid(self):
        """Generates a grid with the coordinates that are occupied by elements given the folds"""
        grid={}
        c=Coordinate(self.dimension)
        grid[str(c)]=self.elements[0]
        for i in range(len(self.folds)):
            c.fold(self.folds[i])
            if str(c) in grid:
                raise ProteinFoldException("Fold nr_"+str(i)+" ("+str(self.folds[i])+") causes the protein to fold in on itself")
            grid[str(c)]=self.elements[i+1]
        return grid

    def score_protein_elements(self,element1,element2):
        """Returns the score of two neigboring elements"""
        if element1=='H':
            if element2=='H':
                return 1
        return 0

    def score_coord(self,grid,coord):
        """Computes the score by scoring the element in coord with all elemnts in the coords increased by 1 in all dimension.
           We don't look in de decreased direction to avoid double counting the scores."""
        score=0
        element1=grid[str(coord)]
        for i in range(self.dimension):
            coord.fold(i+1)
            element2=grid.get(str(coord))
            if not element2 is None:
                score+=self.score_protein_elements(element1,element2)
            coord.fold(-(i+1)) # restore coord to previous value
        return score

    def score_grid(self,grid):
        """Walks through the grid and scores each neigboring coordinate"""
        score=0
        c=Coordinate(self.dimension)
        score+=self.score_coord(grid,c)
        for i in range(len(self.folds)):
            c.fold(self.folds[i])
            score+=self.score_coord(grid,c)
        return score

    def score_element_list(self):
        """Walks through "elements" list and scores each neigboring element"""
        score=0
        for i in range(len(self.elements)-1):
            score+=self.score_protein_elements(self.elements[i],self.elements[i+1])
        return score

    def compute_score(self):
        """Computes the score of the Protein"""
        if len(self.folds)==0:
            return 0
        grid=self.make_grid()
        score=self.score_grid(grid) # compute by scoring each coordinate in the grid
        score-=self.score_element_list() # then subtract the score of all neigboring elements in the "elements" list
        return score
